/*
*
* Copyright (C) 2007 Mike Blamires
* All Rights Reserved
* Released under GPL2.0 - please see attached document.
*
* http://blog.blamires.co.uk/
* mike@blamires.co.uk
*
*/

using System;
using System.Collections.Generic;
using System.Text;

using System.Collections;
using System.IO;
using System.Net.Sockets;

using Pop3Protocol.Exceptions;


namespace Pop3Protocol.Protocol {

    /* Author: Mike Blamires
     * Version: 0.1
     * 
     * 
     * Description:
     * RFC 1969 Compliant POP3 Client class.
     * 
     * */
      
    public class Pop3 {

        private TcpClient tcpClient;
        private NetworkStream ns;
        private StreamReader sr;
        private StreamWriter sw;

        private Boolean isConnected = false;

        public Pop3() {

        }

        public Pop3(String server, int port, String username, String password) {
            this.connect(server, port, username, password);
        }

        public Boolean connect (String server, int port, String username, String password) {
            tcpClient = new TcpClient(server, port);
            ns = tcpClient.GetStream();
            sw = new StreamWriter(ns);
            sr = new StreamReader(ns);
            sw.AutoFlush = true;

            if (ns.CanRead && ns.CanWrite) {
                if (tcpClient.Connected) {
                    sr.ReadLine();

                    sw.WriteLine(String.Format("USER {0}", username));

                    String firstResponse = sr.ReadLine();
                      
                    if (firstResponse.StartsWith("+OK")) {
                        sw.WriteLine(String.Format("PASS {0}", password));
                        Console.WriteLine(String.Format("PASS {0}", password));
                        firstResponse = sr.ReadLine();
                        if (firstResponse.StartsWith("+OK")) {
                            this.isConnected = true;
                            return true;
                        }
                    }
                }
            } else {
                return false;
            }
            return false;
        }

        public Boolean disconnect() {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {
                    sw.WriteLine("QUIT");
                    tcpClient.Close();
                    this.isConnected = false;
                    return true;
                }
            } //no else, if not connected we don't need to disconnect
            return false; //that would be weird!
        }

        /// <summary>
        /// returns the result of the stat account displaying a count and mailbox size in a String array
        /// </summary>
        /// <returns>String</returns>
        public String[] stat() {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine("STAT");

                    String firstResponse = sr.ReadLine();

                    if (firstResponse.StartsWith("+OK")) {
                        String[] response = firstResponse.Split(' ');
                        return new String[] { response[1].ToString(), response[2].ToString() };
                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }

                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        /// <summary>
        /// Returns a Key & Value pair Dictionary of the messages and their size in bytes.
        /// This does not reference messages marked for deletion.
        /// If there is no 
        /// </summary>
        /// <returns>Dictionary[int,int]</int></returns>
        public Dictionary<int, int> list() {
            Dictionary<int, int> dictionary = new Dictionary<int, int>();
            if (IsConnected) {

                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine("LIST");

                    String firstResponse = sr.ReadLine(); 
                    if (firstResponse.StartsWith("+OK")) {
                       
                        String response;
                        String[] line_s;
                        while ((response = sr.ReadLine()) != ".") {
                            line_s = response.Split(' ');
                            dictionary.Add(int.Parse(line_s[0]), int.Parse(line_s[1]));
                        }                       

                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }
                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
            return dictionary;
        }

        /// <summary>
        /// Gets the size in bytes of the email parsed from the servers return, n cannot be 
        /// a message marked for deletion.
        /// </summary>
        /// <param name="n">Message number</param>
        /// <returns>int</returns>
        public int list(int n) {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine(String.Format("LIST {0}", n.ToString()));

                    String response = sr.ReadLine();

                    if (response.StartsWith("+OK")) {

                        String[] line_s = response.Split(' ');
                        return int.Parse(line_s[2]);

                    } else {
                        if (response.Length > 4) {
                            throw new ServerErrorException(response);
                        } else {
                            throw new ServerErrorException();
                        }
                    }

                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        /// <summary>
        /// Gets the whole email (n) as plain text return for parsing by the client.
        /// </summary>
        /// <param name="n">the email to retrieve</param>
        /// <returns>String</returns>
        public String retr(int n) {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine(String.Format("RETR {0}", n.ToString()));

                    String firstResponse = sr.ReadLine();

                    if (firstResponse.StartsWith("+OK")) {
                        String line;
                        String email = "";
                        while ((line = sr.ReadLine()) != ".") {
                            email += line + "\r\n";
                        }
                        return email;
                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }
                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        public Boolean dele(int n) {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine(String.Format("DELE {0}", n.ToString()));

                    String firstResponse = sr.ReadLine();

                    if (firstResponse.StartsWith("+OK")) {
                        return true;
                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }

                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        /// <summary>
        /// Sends a NOOP command. No result to process.
        /// </summary>
        /// <returns>Boolean</returns>
        public Boolean noop() {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine("NOOP");

                    String response = sr.ReadLine();
                                
                    if (response.StartsWith("+OK")) {
                        return true;
                    } else {
                        if (response.Length > 4) {
                            throw new ServerErrorException(response);
                        } else {
                            throw new ServerErrorException();
                        }
                    }

                } else {
                    throw new StreamNotAvailableException();     
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        /// <summary>
        /// Resets the mailbox, so any items deleted in the current transaction state are reinstated
        /// </summary>
        /// <returns>Boolean</returns>
        public Boolean rset() {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine("RSET");

                    String response = sr.ReadLine();

                    if (response.StartsWith("+OK")) {
                        return true;
                    } else {
                        return false;
                    }

                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }



        /* Optional commands.
         * TOP n x
         * UIDL 
         * APOP is not at this point supported.
         */

        /// <summary>
        /// Returns the number of lines specified by lines, in message n. if l is bigger than the
        /// total lines in teh email the whole email will return.
        /// </summary>
        /// <param name="n">The email to be retrieved</param>
        /// <param name="lines">The number of lines to parse (including header)</param>
        /// <returns>String</returns>
        public String top(int n, int lines) {
            if (IsConnected) {
                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine(String.Format("TOP {0} {1}", n.ToString(), lines.ToString()));

                    String firstResponse = sr.ReadLine();

                    if (firstResponse.StartsWith("+OK")) {
                        String line;
                        String email = "";
                        while ((line = sr.ReadLine()) != ".") {
                            email += line + "/r/n";
                        }
                        return email;
                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }
                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
        }

        /// <summary>
        /// Returns a list of the unique IDs of all the mail in the mailbox
        /// </summary>
        /// <returns>Dictionary[int,string]</returns>
        public Dictionary<int, string> uidl() {
            Dictionary<int, string> dictionary = new Dictionary<int, string>();
            if (IsConnected) {                

                if (ns.CanRead && ns.CanWrite) {

                    sw.WriteLine("UIDL");
                    String firstResponse = sr.ReadLine(); 
                    if (firstResponse.StartsWith("+OK")) {
                       
                        String response;
                        String[] line_s;
                        while ((response = sr.ReadLine()) != ".") {
                            line_s = response.Split(' ');
                            dictionary.Add(int.Parse(line_s[0]), line_s[1]);
                        }     

                    } else {
                        if (firstResponse.Length > 4) {
                            throw new ServerErrorException(firstResponse);
                        } else {
                            throw new ServerErrorException();
                        }
                    }
                } else {
                    throw new StreamNotAvailableException();
                }
            } else {
                throw new ClientNotConnectedException();
            }
            return dictionary;
        }


        /// <summary>
        /// Returns if the communications stream is in the TRANSACTION state
        /// </summary>
        public Boolean IsConnected {
            get {
                return isConnected;
            }
        }


    }
            
}
